/*
    learn.cpp
    $Id: learn.cpp,v 1.7 2014/12/19 02:38:59 daichi Exp $

*/
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <algorithm>
#include "learn.h"
#include "random.h"
#include "id.h"
#include "util.h"
#define	 INTERVAL	100

using namespace std;

void
predict_model (vector<double> &y, vector<vector<int> > &data,
	       vector<double> &wa, vector<double> &wb)
{
	double a, b;
	int n, ndata = data.size();

	for (n = 0; n < ndata; n++)
	{
		a = exp (spdot(wa, data[n]));
		b = exp (spdot(wb, data[n]));
		printf("%lg\t%.6lf\t%.6lf\n", y[n], a, b);
	}
}

void
learn_model (vector<double> &y, vector<vector<int> > &data,
	     vector<double> &wa, vector<double> &wb,
	     int iters, double eps, double sigma)
{
	static vector<int> seq;
	static vector<vector<int> > indata;
	int i, j, m, n, t, f, F = ids() + 1;
	int len = 2 * F, ndata = data.size();
	int accept;
	double lik;

	/* init */
	wa.resize (F);
	wb.resize (F);
	seq.resize (len);
	for (i = 0; i < len; i++)
		seq[i] = i;

	/* prepare */
	indata.resize (F);
	for (n = 0; n < ndata; n++)
	{
		for (j = 0; j < data[n].size(); j++)
		{
			f = data[n][j] + 1; // + bias term
			indata[f].push_back (n);
		}
	}

	printf("features = %d, ndata = %d\n", F, ndata);

	/* MCMC */
	for (m = 0; m < iters; m++)
	{
		accept = 0;
		random_shuffle (seq.begin(), seq.end(), irand);

		for (t = 0; t < len; t++)
		{
			if (((t + 1) % INTERVAL) == 0) {
				fprintf(stderr, "feature %4d/%d..\r", t+1, len);
				fflush(stderr);
			}
			if (seq[t] < F) {
				f = seq[t];
				accept += mh_update (wa, f, wa, wb, y, data,
						     indata, eps, sigma);
			} else {
				f = seq[t] - F;
				accept += mh_update (wb, f, wa, wb, y, data,
						     indata, eps, sigma);
			}
		}
		printf("iter[%2d] : PPL = %lf (accept = %.2f%%)\n",
		       m + 1, exp (- gamglmlik (y, data, wa, wb) / ndata),
		       (double) accept/seq.size() * 100);
		
	}
		
}

int
mh_update (vector<double> &w, int f,
	   vector<double> &wa, vector<double> &wb,
	   vector<double> &y, vector<vector<int> > &data,
	   vector<vector<int> > &indata, double eps, double sigma)
{
	double pw = w[f];
	double lik0, lik1;

	lik0 = gamglmflik (f, y, data, indata, wa, wb) + lnorm(w[f], sigma);
	w[f] += eps * NRANDOM;
	lik1 = gamglmflik (f, y, data, indata, wa, wb) + lnorm(w[f], sigma);

	if (bernoulli(exp(lik1 - lik0)))
	{
		return 1;	/* accept */
	} else {
		w[f] = pw;	/* revert */
		return 0;
	}
}

double
gamglmlik (vector<double> &y, vector<vector<int> > &data,
	   vector<double> &wa, vector<double> &wb)
{
	int n, ndata = data.size();
	double lik = 0;

	for (n = 0; n < ndata; n++)
	{
		lik += gamreglik (y[n], data[n], wa, wb);
	}

	return lik;
}

double
gamglmflik (int f, vector<double> &y, vector<vector<int> > &data,
	   vector<vector<int> > &indata,
	   vector<double> &wa, vector<double> &wb)
{
	int i, n, N = indata[f].size();
	double lik = 0;

	if (f == 0)
	{
		lik = gamglmlik (y, data, wa, wb);
	}
	else {
		for (i = 0; i < N; i++)
		{
			n = indata[f][i];
			lik += gamreglik (y[n], data[n], wa, wb);
		}
	}

	return lik;
}

double
gamreglik (double y, vector<int> &data,
	   vector<double> &wa, vector<double> &wb)
{
	double a = exp (spdot(wa, data));
	double b = exp (spdot(wb, data));

	return (a * log(b) - lgamma(a) + (a - 1) * log(y) - b * y);
}

double
spdot (vector<double> &w, vector<int> &v)	/* sparse dot */
{
	double z = w[0];
	int f, i, n = v.size();

	for (i = 0; i < n; i++)
	{
		f = v[i] + 1;	// plus bias term
		z += w[f];	// value is always 1
	}

	return z;
}

double
lnorm (double w, double sigma)
{
	return - (w * w) / (2 * sigma * sigma);
}

void
save_vector (char *file, vector<double> &v)
{
	FILE *fp;
	int i, n = v.size();

	if ((fp = fopen(file, "w")) == NULL) {
		error("save_vector: cannot open %s to write.", file);
		return;
	}
	
	for (i = 0; i < n; i++)
	{
		fprintf(fp, "% .7e\n", v[i]);
	}

	fclose(fp);
}

void
load_vector (char *file, vector<double> &v)
{
	FILE *fp;
	double val;
	char line [BUFSIZ];

	if ((fp = fopen(file, "r")) == NULL) {
		error("load_vector: cannot open %s to read.", file);
		return;
	}

	while (fgets(line, sizeof(line), fp))
	{
		val = atof(line);
		v.push_back (val);
	}

	fclose(fp);
}
			
	       

void
save_model (char *file, vector<double> &wa, vector<double> &wb)
{
	save_vector (strconcat(file, ".a"), wa);
	save_vector (strconcat(file, ".b"), wb);
	save_dict (file);
}

void
load_model (char *file, vector<double> &wa, vector<double> &wb)
{
	load_vector (strconcat(file, ".a"), wa);
	load_vector (strconcat(file, ".b"), wb);
	load_dict (file);
}


